﻿using Newtonsoft.Json;

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Timers;
using System.Windows.Forms;

namespace TestingPosScale
{
    public partial class Form1 : Form
    {
        private readonly List<int> _baudRates = new() { 1200, 2400, 9600, 19200 };
        private readonly List<int> _dataBits = new() { 5, 6, 7, 8 };
        private readonly List<int> _stopBits = new() { 0, 1, 2 };
        private readonly System.Timers.Timer _sendTimer;
        private SerialPort _port;
        private delegate void SetResponseTextCallback(string responseText);

        public Form1()
        {
            InitializeComponent();

            cmbo_Port.DataSource = SerialPort.GetPortNames();

            cmbo_Baud.DataSource = _baudRates;
            cmbo_Baud.SelectedIndex = cmbo_Baud.FindStringExact("9600");

            cmbo_DataBits.DataSource = _dataBits;
            cmbo_DataBits.SelectedIndex = cmbo_DataBits.FindStringExact("7");

            cmbo_StopBits.DataSource = _stopBits;
            cmbo_StopBits.SelectedIndex = cmbo_StopBits.FindStringExact("1");

            rbtn_ParityEven.Checked = true;

            tb_InitString.Text = "W";

            _sendTimer = new System.Timers.Timer(1000) { AutoReset = false };
            _sendTimer.Elapsed += new ElapsedEventHandler(InputTimer_Tick);
        }

        private void InputTimer_Tick(object sender, ElapsedEventArgs e)
        {
            var s = GetInitString();
            Logger.Log(Logger.Level.Debug, $"Sent: [{s}]");

            _port.Write(s);
        }

        private void Btn_Start_Click(object sender, EventArgs e)
        {
            var errors = ValidateOptions();

            if (errors.Count > 0)
            {
                MessageBox.Show(string.Join(Environment.NewLine, errors), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);

                return;
            }

            var port = cmbo_Port.SelectedValue.ToString();
            var baud = Convert.ToInt32(cmbo_Baud.SelectedValue);
            var dataBits = Convert.ToInt32(cmbo_DataBits.SelectedValue);
            var stopBits = (StopBits)Convert.ToInt32(cmbo_StopBits.SelectedValue);
            var parity = GetParity();

            DisableControls();

            try
            {
                if (cb_EmulatePOS.Checked)
                {
                    _port = new SerialPort(port);
                }
                else
                {
                    _port = new SerialPort(port, baud, parity, dataBits, stopBits)
                    {
                        DtrEnable = true,
                        Handshake = Handshake.RequestToSend
                    };
                }

                _port.DataReceived += new SerialDataReceivedEventHandler(SerialDataReceivedHandler);
                _port.Open();

                if (cb_EmulatePOS.Checked)
                {
                    _port.DiscardOutBuffer();
                    _port.DiscardInBuffer();
                    _port.RtsEnable = true;
                }

                var config = new
                {
                    Port = port,
                    Baud = baud,
                    DataBits = dataBits,
                    StopBits = stopBits.ToString(),
                    Parity = parity.ToString(),
                    InitString = tb_InitString.Text,
                    SendCR = cb_SendCR.Checked,
                    SendLF = cb_SendLF.Checked,
                    SendCRLF = cb_SendCRLF.Checked,
                    EmulatePos = cb_EmulatePOS.Checked
                };

                Logger.Log(Logger.Level.Debug, JsonConvert.SerializeObject(config));

                _sendTimer.Start();
            }
            catch (Exception ex)
            {
                Logger.Log(Logger.Level.Debug, ex);

                CloseAndDisposePort();

                MessageBox.Show(ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);

                EnableControls();
            }
        }

        private void Btn_Stop_Click(object sender, EventArgs e)
        {
            _sendTimer.Stop();
            
            CloseAndDisposePort();

            EnableControls();
        }

        private string GetInitString()
        {
            if (cb_SendCR.Checked)
            {
                return $"{tb_InitString.Text}\r";
            }
            else if (cb_SendLF.Checked)
            {
                return $"{tb_InitString.Text}\n";
            }
            else if (cb_SendCRLF.Checked)
            {
                return $"{tb_InitString.Text}\r\n";
            }

            return tb_InitString.Text;
        }

        private List<string> ValidateOptions()
        {
            var errors = new List<string>();

            if (cmbo_Port.SelectedIndex < 0)
            {
                errors.Add("No COM port selected.");
            }

            if (cmbo_Baud.SelectedIndex < 0)
            {
                errors.Add("No baud rate selected.");
            }

            if (cmbo_DataBits.SelectedIndex < 0)
            {
                errors.Add("No data bits selected.");
            }

            if (cmbo_StopBits.SelectedIndex < 0)
            {
                errors.Add("No stop bits selected.");
            }

            if (!rbtn_ParityEven.Checked
                && !rbtn_ParityOdd.Checked
                && !rbtn_ParityMark.Checked
                && !rbtn_ParitySpace.Checked
                && !rbtn_ParityNone.Checked)
            {
                errors.Add("No parity selected.");
            }

            if (string.IsNullOrWhiteSpace(tb_InitString.Text))
            {
                errors.Add("No init string specified.");
            }

            return errors;
        }

        private Parity GetParity()
        {
            var parity = Parity.None;

            if (rbtn_ParityEven.Checked)
            {
                parity = Parity.Even;
            }
            else if (rbtn_ParityOdd.Checked)
            {
                parity = Parity.Odd;
            }
            else if (rbtn_ParityMark.Checked)
            {
                parity = Parity.Mark;
            }
            else if (rbtn_ParitySpace.Checked)
            {
                parity = Parity.Space;
            }

            return parity;
        }

        private void SerialDataReceivedHandler(object sender, SerialDataReceivedEventArgs e)
        {
            var stringData = _port.ReadExisting();
            var data = new byte[_port.BytesToRead];
            _port.Read(data, 0, data.Length);

            Logger.Log(Logger.Level.Debug, $"Decoded bytes: [{Encoding.UTF8.GetString(data)}]");

            SetResponseText(stringData);

            var sb = new StringBuilder();
            var chars = new List<int>();

            foreach (var c in stringData)
            {
                chars.Add(c);

                if (c == (char)2 || c == (char)10 || c == (char)13)
                {
                    continue;
                }

                sb.Append(c);
            }

            Logger.Log(Logger.Level.Debug, $"SB: [{sb}]");

            Logger.Log(Logger.Level.Debug, string.Join(", ", chars));

            _sendTimer?.Start();
        }

        private void SetResponseText(string responseText)
        {
            if (tb_Response.InvokeRequired)
            {
                var d = new SetResponseTextCallback(SetResponseText);
                Invoke(d, new object[] { responseText });
            }
            else
            {
                string[] lines = responseText.Split(new[] { Environment.NewLine }, StringSplitOptions.None);

                foreach (var line in lines)
                {
                    tb_Response.AppendText($"[{DateTime.Now:yyyy-MM-dd HH:mm:ss.fff}]  {line}{Environment.NewLine}");
                    Logger.Log(Logger.Level.Debug, $"[{responseText}]");
                }
            }
        }

        private void CloseAndDisposePort()
        {
            if (_port is not null && _port.IsOpen)
            {
                _port.DataReceived -= new SerialDataReceivedEventHandler(SerialDataReceivedHandler);
                _port.Close();
                _port = null;
            }
        }

        private void DisableControls()
        {
            cmbo_Port.Enabled = false;
            cmbo_Baud.Enabled = false;
            cmbo_DataBits.Enabled = false;
            cmbo_StopBits.Enabled = false;
            rbtn_ParityEven.Enabled = false;
            rbtn_ParityOdd.Enabled = false;
            rbtn_ParityMark.Enabled = false;
            rbtn_ParitySpace.Enabled = false;
            rbtn_ParityNone.Enabled = false;
            tb_InitString.Enabled = false;
            cb_SendCR.Enabled = false;
            cb_SendLF.Enabled = false;
            cb_SendCRLF.Enabled = false;
            cb_EmulatePOS.Enabled = false;
            btn_Start.Enabled = false;
        }

        private void EnableControls()
        {
            cmbo_Port.Enabled = true;
            cmbo_Baud.Enabled = true;
            cmbo_DataBits.Enabled = true;
            cmbo_StopBits.Enabled = true;
            rbtn_ParityEven.Enabled = true;
            rbtn_ParityOdd.Enabled = true;
            rbtn_ParityMark.Enabled = true;
            rbtn_ParitySpace.Enabled = true;
            rbtn_ParityNone.Enabled = true;
            tb_InitString.Enabled = true;
            cb_SendCR.Enabled = true;
            cb_SendLF.Enabled = true;
            cb_SendCRLF.Enabled = true;
            cb_EmulatePOS.Enabled = true;
            btn_Start.Enabled = true;
        }

        private void Cb_Send_CheckedChanged(object sender, EventArgs e)
        {
            if (sender is CheckBox cb)
            {
                switch (cb.Name)
                {
                    case nameof(cb_SendCR):
                        if (cb.Checked)
                        {
                            cb_SendLF.Checked = false;
                            cb_SendCRLF.Checked = false;
                        }
                        break;

                    case nameof(cb_SendLF):
                        if (cb.Checked)
                        {
                            cb_SendCR.Checked = false;
                            cb_SendCRLF.Checked = false;
                        }
                        break;

                    case nameof(cb_SendCRLF):
                        if (cb.Checked)
                        {
                            cb_SendCR.Checked = false;
                            cb_SendLF.Checked = false;
                        }
                        break;
                }
            }
        }
    }
}
